
#include "PinChangeInterrupt.h"

#include <OSCBundle.h>
#include <OSCBoards.h>

#ifdef BOARD_HAS_USB_SERIAL
#include <SLIPEncodedUSBSerial.h>
SLIPEncodedUSBSerial SLIPSerial( thisBoardsSerialUSB );
#else
#include <SLIPEncodedSerial.h>
SLIPEncodedSerial SLIPSerial(Serial);
#endif

// Motor CTRL "Library"
#include "motor_ctrl.h"

// Tuning PID
#include <PID_v1.h>

motor_ctrl myMotor_0(3, 5, 11, 16 );
motor_ctrl myMotor_1(6, 9, 14, 15 );

// Send position time guard , not too fast
unsigned long fps_ts;
const int fps_interval = 100;

// Motor stop frames
const int motor_frames_interval = 3000;
unsigned long motor_stopped[2];

void setup() {
  // Wait for motor to settle , need to implement self test later
  delay(1000);

  // Init motor
  myMotor_0.init();
  myMotor_1.init();
  // This needed for failure to register inside class to instance member function
  attachPCINT(digitalPinToPCINT(myMotor_0.get_sensorPin(0)), myMotor_0_Toggle0, CHANGE);
  attachPCINT(digitalPinToPCINT(myMotor_0.get_sensorPin(1)), myMotor_0_Toggle1, CHANGE);

  attachPCINT(digitalPinToPCINT(myMotor_1.get_sensorPin(0)), myMotor_1_Toggle0, CHANGE);
  attachPCINT(digitalPinToPCINT(myMotor_1.get_sensorPin(1)), myMotor_1_Toggle1, CHANGE);

  myMotor_0.SetSampleTime(10);  // 50 fps
  myMotor_0.SetControllerDirection(REVERSE);  // PID library bug (always re set tuning AFTER setting control direction)
  myMotor_0.SetTunings(1.4, 3.75, 0.2);

  myMotor_1.SetSampleTime(10);  // 50 fps
  myMotor_1.SetControllerDirection(REVERSE);  // PID library bug (always re set tuning AFTER setting control direction)
  myMotor_1.SetTunings(1.4, 3.75, 0.2);
  // Set mode last
  myMotor_0.SetMode(AUTOMATIC);
  myMotor_1.SetMode(AUTOMATIC);

  // For output test
  fps_ts = millis();

  // SLIP begin
  SLIPSerial.begin(115200);   // set this as high as you can reliably run on your platform
#if ARDUINO >= 100
  while (!Serial)
    ;   // Leonardo bug
#endif

  // Serial Load complete
  delay(1000);
  OSCBundle oscRdy;

  oscRdy.add("/rdy");
  SLIPSerial.beginPacket();
  oscRdy.send(SLIPSerial);
  SLIPSerial.endPacket();
  oscRdy.empty();
}

void loop() {
  // Send current position
  if ( millis() - fps_ts >= fps_interval ) {
    if ( myMotor_0.moved() || myMotor_1.moved() ) {
      if( myMotor_0.moved() ){
        motor_stopped[0] = millis();
      }else{
        motor_stopped[1] = millis();
      }
      fps_ts = millis();
      OSCBundle oscMtr;

      oscMtr.add("/mtr").add((int32_t)myMotor_0.get_currentPos()).add((int32_t)myMotor_1.get_currentPos());
      SLIPSerial.beginPacket();
      oscMtr.send(SLIPSerial);
      SLIPSerial.endPacket();
      oscMtr.empty();
    }

    if( millis() - motor_stopped[0] > motor_frames_interval ){
      myMotor_0.stop_motor();
    }
    if( millis() - motor_stopped[1] > motor_frames_interval ){
      myMotor_1.stop_motor();
    }
  }
  // Receive data
  OSCBundle bundleIN;
  int size;

  if ( SLIPSerial.available() ) {
    while (!SLIPSerial.endofPacket())
      if ( (size = SLIPSerial.available()) > 0)
      {
        while (size--)
          bundleIN.fill(SLIPSerial.read());
      }

    if (!bundleIN.hasError()) {
      bundleIN.dispatch("/mtr", setMtr);
      bundleIN.dispatch("/mtrclr", clrMtr);
      bundleIN.dispatch("/getset", getSet);
    }
  }

  // Run motor
  myMotor_0.process_motor();
  myMotor_1.process_motor();
}

void setMtr(OSCMessage &msg) {
  if ( msg.isInt(0) && msg.isInt(1)) {
    if ( msg.getInt(0) == 0 ) {
      myMotor_0.set_position(msg.getInt(1));
    } else {
      myMotor_1.set_position(msg.getInt(1));
    }
  }
}

void clrMtr(OSCMessage &msg) {
  if ( msg.isInt(0) ) {
    if ( msg.getInt(0) == 0 ) {
      myMotor_0.clear_pos();
    } else {
      myMotor_1.clear_pos();
    }
  }
}

// Get setpoint for debug
void getSet(OSCMessage &msg) {
  OSCBundle oscMtr;
  oscMtr.add("/mtrset").add((int32_t)myMotor_0.get_setPos()).add((int32_t)myMotor_1.get_setPos());
  SLIPSerial.beginPacket();
  oscMtr.send(SLIPSerial);
  SLIPSerial.endPacket();
  oscMtr.empty();
}

// Motor pin change Interrupt service routine
void myMotor_0_Toggle0(void) {
  myMotor_0.togglePin0();
}

void myMotor_0_Toggle1(void) {
  myMotor_0.togglePin1();
}

void myMotor_1_Toggle0(void) {
  myMotor_1.togglePin0();
}

void myMotor_1_Toggle1(void) {
  myMotor_1.togglePin1();
}
// Motor pin change Interrupt service routine ends

