#ifndef MOTOR_CTRL_CPP_
#define MOTOR_CTRL_CPP_

#include "motor_ctrl.h"

const int8_t motor_ctrl::ms_trans_matrix[4][4] = {
  {0, 1, -1, 0},
  { -1, 0, 0, 1},
  {1, 0, 0, -1},
  {0, -1, 1, 0}
};

motor_ctrl::motor_ctrl( int8_t pin1, int8_t pin2 , int8_t sensorPin1, int8_t sensorPin2 )
  : m_SensorPins( {
  sensorPin1, sensorPin2
}),
m_Pins({pin1, pin2}),
m_mtr_pos(0),
m_sensor_val(0),
m_sensor_val_last(0),
m_mode(MTR_CTRL_POS), // Default control position, set to 1 to control speed
m_motor_diff(0.0f),
m_set_point(0.0f),
m_set_pos(0),
m_moved(false),
m_PID_ctrl(new PID(&m_motor_diff, &m_motor_out_val, &m_set_point, 2.0, 0.5, 2.0, DIRECT))
{
  m_PID_ctrl->SetOutputLimits(-255, 255);

  // get initial state
  m_sensor_val |= digitalRead( m_SensorPins[0] );
  m_sensor_val != digitalRead( m_SensorPins[1] ) << 1;
}

motor_ctrl::motor_ctrl() {

}

motor_ctrl::~motor_ctrl() {
  free(m_PID_ctrl); // Prevent from possible memory leak
}

void motor_ctrl::init() {
  for ( int i = 0; i < 2; i++ ) {
    pinMode(m_Pins[i], OUTPUT);
    analogWrite( m_Pins[i], 0 );  // make sure stop
    pinMode(m_SensorPins[i], INPUT);
  }
}

// Route PID function out
void motor_ctrl::SetSampleTime(int st) {
  m_PID_ctrl->SetSampleTime(st);
}
void motor_ctrl::SetTunings(double Kp, double Ki, double Kd) {
  m_PID_ctrl->SetTunings(Kp, Ki, Kd);
}
void motor_ctrl::SetMode(int Mode) {
  m_PID_ctrl->SetMode(Mode);
}
void motor_ctrl::SetControllerDirection(int Dir) {
  m_PID_ctrl->SetControllerDirection(Dir);
}

// Motor process (Needs to be called in loop)
void motor_ctrl::process_motor(void) {  // speed not implemeted yet
  switch ( m_mode ) {
    case MTR_CTRL_POS:
      int64_t t_diff = m_mtr_pos - m_set_pos;
      m_motor_diff = t_diff; // give position to input
      break;
  }

  if ( m_PID_ctrl->Compute() ) { // PID output is present
    _set_motor_output_();
  }
}

void motor_ctrl::set_point(double setpoint) { // Set setpoint
  m_set_point = setpoint;
}

void motor_ctrl::set_position( unsigned long setpos ) {
  m_set_pos = setpos;
}

boolean motor_ctrl::moved() {
  boolean result = m_moved;
  m_moved = false;  // clear flag
  return result;
}

// Fuzzy stop
// For outside to call when the motor has stopped for a moment
// But still has a little bit of output
void motor_ctrl::stop_motor() {
  if (m_motor_diff == 0) {
    m_motor_out_val = 0;
    _set_motor_output_();
  }
}

// Since the failure of attaching interrupt to instance member in init, we need to get sensor pin to manually call toggle pin from outside ISR
int8_t motor_ctrl::get_sensorPin( int8_t id ) {
  return m_SensorPins[id];
}

// Get the current position of motor
int32_t motor_ctrl::get_currentPos() {
  return m_mtr_pos;
}

int32_t motor_ctrl::get_setPos() {
  return m_set_pos;
}

void motor_ctrl::togglePin0(void) {
  // Toggle first bit
  m_sensor_val &= 0b11111110;
  m_sensor_val |= digitalRead(m_SensorPins[0]);
  //  m_sensor_val ^= 0b00000001;

  _sense_motor_pos_();
}

void motor_ctrl::togglePin1(void) {
  // Toggle second bit
  m_sensor_val &= 0b11111101;
  m_sensor_val |= digitalRead(m_SensorPins[1]) << 1;
  //  m_sensor_val ^= 0b00000010;

  _sense_motor_pos_();
}

void motor_ctrl::clear_pos() {
  // Force stop
  m_motor_out_val = 0;
  m_set_pos = 0;

  // Clear output
  m_motor_out_val = 0;
  _set_motor_output_();

  // Clear position
  m_mtr_pos = 0;
  m_sensor_val = 0;

  m_sensor_val |= digitalRead( m_SensorPins[0] );
  m_sensor_val != digitalRead( m_SensorPins[1] ) << 1;

  m_sensor_val_last = m_sensor_val;
}

// internal function, add counter
inline void motor_ctrl::_sense_motor_pos_() {
  m_mtr_pos += ms_trans_matrix[m_sensor_val_last][m_sensor_val];
  m_sensor_val_last = m_sensor_val;
  m_moved = true;
}

inline void motor_ctrl::_set_motor_output_() {
  analogWrite( m_Pins[0], m_motor_out_val > 0 ? m_motor_out_val : 0 );
  analogWrite( m_Pins[1], m_motor_out_val < 0 ? -m_motor_out_val : 0 );
}

#endif // MOTOR_CTRL_CPP_
